#!/usr/bin/Rscript
#author : olivier.filangi@rennes.inra.fr
#
# ARG1 => Variance Components Estimation file provided by GS3 (http://snp.toulouse.inra.fr/~alegarra/)
# ARG2 => OUTPUT File in HTML format
#
#
#librairie CODA
library(coda)

args <- commandArgs(TRUE)

if (length(args) != 2 ) {
cat("argument missing :1 variance components sample gs3 file, 2 output.html!\n")
stop()
}

#le fichier variance components sample
mcmc.gs3 <-read.table(args[1],header=TRUE)

#output in HTML format
sink(args[2], append=FALSE, split=FALSE)

res<-mcmc(mcmc.gs3)

cat('<p>')
packinfo <- installed.packages (fields = c ("Package", "Version"))
cat ('This script use CODA Library [',packinfo["coda", "Version"],'] <b>Output analysis and diagnostics for Markov Chain Monte Carlo simulations.</b>\n<br/>')
cat ('<a href="http://cran.r-project.org/web/packages/coda/coda.pdf">Reference manual of CODA</a>\n')
cat('<p>\n')

#--------- SUMMARY ------------

cat ("<h2>Summary</h2>\n")
cat ("<pre>\n")
summary(res)
cat ("</pre>\n")

# print a traceplot (on each  iteration => the value of the var. we can see stuck if bad mixing)
# print denseplot 
cat ("<h3>Trace and Density</h3>\n")

cat ("<table>")

for (i in 1:ncol(res)) {
 #file name
 namePngFile<-paste(colnames(res)[i],'.png')
 #remove blank
 namePngFile<-gsub(" ","",namePngFile, fixed=TRUE)
 sdata <- as.matrix(mcmc.gs3[,i])
 sthin=1
 send<- nrow(sdata)
 sstart=1
 attr(sdata, "mcpar") <- c(sstart, send, sthin)
 attr(sdata, "class") <- "mcmc"
 attr(sdata, "dimnames") <- list(NULL,c(colnames(res)[i]))
 #manage png generation
 png(file=namePngFile,width=800,res=110)
 #plot
 plot(sdata)
 #end of png
 invisible(dev.off())
 # generation of a string for web image embedded format 
 pngString<-system(paste('base64 -w 0 ',namePngFile),intern=TRUE)
 pngString<-paste(pngString, collapse = "")
 cat ("<tr><td>")
 cat('<img height="400" alt="Trace and Density of ',colnames(res)[i],'" src="data:image/png;base64,',pngString,'"\"/>\n"')
 cat ("</td></tr>")
 #remove png file
 system(paste('rm -rf ',namePngFile))
}
cat ("</table>")

#print running mean plot
namePngFile<-"running_mean_plot.png"
png(file=namePngFile,width=800,res=110)
plot.ts(mcmc.gs3)
invisible(dev.off())
pngString<-system(paste('base64 -w 0 ',namePngFile),intern=TRUE)
pngString<-paste(pngString, collapse = "")
cat ("<h3>Running Mean Plot</h3>\n")
cat('<img height="500" alt="Running Mean Plot" src="data:image/png;base64,',pngString,'"\"/>\n"')
system(paste('rm -rf ',namePngFile))

#autocorrelation
namePngFile<-"autocorrelations.png"
png(file=namePngFile,width=800,res=110)
autocorr.plot(res)
invisible(dev.off())
pngString<-system(paste('base64 -w 0 ',namePngFile),intern=TRUE)
pngString<-paste(pngString, collapse = "")
cat ("<h3>Autocorrelations</h3>\n")
cat ('<p><a href="http://en.wikipedia.org/wiki/Autocorrelation#Statistics">description</a></p>\n')

cat('<img height="500" alt="Autocorrelations" src="data:image/png;base64,',pngString,'"\"/>\n"')
system(paste('rm -rf ',namePngFile))


#sink(args[2], append=TRUE, split=FALSE)
#Geweke diagnostic 
cat ("<h3>Geweke diagnostic</h3>\n")

cat ("<pre>\n")
geweke.diag(res)
cat ("</pre>\n")

#heidelberg and welch diagnostic
cat ("<h3>Heidelberg and Welch diagnostic</h3>\n")
cat ("<pre>\n")
heidel.diag(res)
cat ("</pre>\n")


